/* (This is the new BSD license.)
* Copyright (c) 2014-2016, Chris Culy
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the Chris Culy nor the 
*		names of its contributors may be used to endorse or promote 
*		products from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY Chris Culy
* ``AS IS'' AND ANY OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
* THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL Chris Culy
* BE LIABLE FOR ANY, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
* GOODS OR SERVICES; OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
* CAUSED AND ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR 
* TORT INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
* THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

"use strict";
/**
 @namespace kwicIS
 All of the functionality is in the kwicIS namespace
*/
var kwicIS = kwicIS || {};

(function(){

//the container is either an ID or the HTMl element
/**
  * @class kwicIS.KWIC
  *  This is the class for the KWIC
  * <p>
  * The constructor
  * @param container is the html element that will contain the KWIC. An id can be used instead.
  */
kwicIS.KWIC = function(container) {
    var where; //the html element for the KWIC
    var id; //the id of the element; we'll make up one if we need to
    var data; // array of objects: {left: array of (TextInfo) leftContexts, hit:(TextInfo), right: array of (TextInfo) rightContexts, index:idx]}
    var fromObjs = false; //is the input data objects (e.g. TextInfo), or delimited Strings
    //if we have delimited strings as input, we need these
    var fldNames = ["token"];
    var fldDelim = "/";
    var textInfoToString; //toString function for textInfo
    var tmu = textmodel.TextModelUtilities(fldNames, fldDelim, textInfoToString); //for when we get delimited string input
    var filters = {}; // {left: array of filter_functions, hit:filter_function, right: array of filter_functions. a filter_function gets an info object and returns Boolean. The function at index <em>i</em> filters position <em>i + 1</em> away from the root. Default is no filtering (via empty arrays)
    var excludedIDs = d3.set(); //ids excluded by filters
    //
    var numHits;
    var leftContextLen = 5; 
    var rtContextLen = 5;
    var numToShow = 20; //number of hits to show at one time
    var whichPage = 0; //which "page" of hits we are showing
    var lastPage; //the last page; read only
    var sortCol = 0; //i.e. the hit. For now just one column.
    var headerData;
    var headers;
    var cellTitle = function(textInfo, i) {}; //noop
    var colHeaderClick = function(headerContent, col) {}; //noop
    var itemCellClick = function(textInfo, i) {}; //noop
    var rowString = function(textInfoArray, i) {}; //noop
    var rowStringOnRight;
    var haveRowString = false;
    var prefixesOnRight = false;
    
    
    var kwic = this; //a way to keep track of the main object since functions each have their own _this_
    
    //we can pass either an ID or the actual HTML element this way
    if (typeof(container) == "string") {
        where = document.getElementById(container);
        id = container;
    } else {
        where = container; // not ideal -- should check that this is really html element
        id = container.id;
        if (!id) {
            container.id = "kwikContainer33";
            id = container.id;
        }
    }
    
    var table = d3.select(where).append("table").attr("class", "kwicTable");
    table.append("thead").append("tr");
    table.append("tbody");
    
    setupHeaderData();
    
    //end of initialization in constructor
    
    
    ///////////     public functions
    
    
    // setter/getters
    
    /**
     * Getter/setter for the data for the KWIC
     * @param input is the data. It can be <strong>either</strong>: 
     * <p>
     * an array of of arrays [leftContextsArray, hitsArray, rightContextsArray, indexArray], where the i-th elements of the arrays correspond
     * and the contexts and hits are delimited strings (as defined by {@link kwicIS.KWIC#fieldNames} and {@link kwicIS.KWIC#fieldDelim}
     * <p>
     * <strong>or</strong>
     * <p>
     * an array of objects, where each object is of the form: {left: array of leftContexts, hits:element, right: array of rightContexts, index:element},
     * and all the elements of the contexts and the hits are themselves objects, typically with at least a "token" key.
     * <p>
     * If the input is delimited strings, they will be converted internally to #textmodel.TextInfo objects
     * <p>
     * If not specified, the current data is returned
     */
    this.data = function(input) {
        if (!arguments.length) {
            return data;
        }
        
        //autodetect whether the input is strings or objects
        if (typeof(input[0]) === "object" && input[0] instanceof Array) { //cf. http://javascript.crockford.com/remedial.html
            data = [];
            numHits = input[1].length;
            for (var i=0;i<numHits;i++) {
              
              if (! prefixesOnRight) {
                data[i] = {'left': tmu.textArrayToTextInfo(input[0][i]),
                              'hit':  new tmu.TextInfo(input[1][i]),
                              'right':tmu.textArrayToTextInfo(input[2][i]),
                              'index':input[3][i],
                              'id': i
                              };
              } else {
                //need to swap left and right and reverse
                var leftTI = tmu.textArrayToTextInfo(input[2][i]);
                leftTI.reverse();
                var rtTI = tmu.textArrayToTextInfo(input[0][i]);
                rtTI.reverse();
                data[i] = {'left': leftTI,
                              'hit':  new tmu.TextInfo(input[1][i]),
                              'right':rtTI,
                              'index':input[3][i],
                              'id': i
                              };
              }
              
            }
            //data = realArray;
        } else {
            numHits = input.length;
            //data = input;
            data = input.map(function(d,i) {
              d.id = i;
              if (prefixesOnRight) {
                //swap left and right and reverse
                d.left.reverse();
                d.right.reverse();
                var tmp = d.left;
                d.right = d.left;
                d.left = tmp;
              }
              return d;
            });
        }
        
        filters = {};
        excludedIDs = d3.set();
        calcLastPage();
        return kwic;
    }
    
    /**
     * @returns just the <em>ids</em> of the data that satisfies the current filters
     */
    this.filteredIDs = function() {
        //return data.filter(function(d) {
        //    return ! excludedIDs.has(d.id);
        //}).map(function(d) {return d.id});
        
        var what = [];
        for(var i=0,n=data.length;i<n;i++){
            var id = d.id;
            if (! excludedIDs.has(d.id)) {
                what.push(id);
            }
        }
        return what;
    }
    
    /**
     * Getter/setter for the filters
     * @param value is an object: {left: array of filter_functions, hit:filter_function, right: array of filter_functions.
     * A filter_function gets an info object and returns Boolean. The function at index <em>i</em> filters position <em>i + 1</em> away from the root. Default is no filtering (via empty arrays).
     * <p>
     * Note: this does <strong>not</strong> redraw the KWIC. Use {@link kwicIS.KWIC#show} for that.
     */
    this.filters = function(value) {
        if (!arguments.length) {
            return filters;
        }
        filters = value;
        filter();
        return kwic;
    }
    
    /**
     * Getter/setter for the field names when the input is delimited strings
     * @param value is an array of strings &mdash; the field names
     */
    this.fieldNames = function(value) {
        if (!arguments.length) {
            return fldNames;
        }
        fldNames = value;
        tmu = textmodel.TextModelUtilities(fldNames, fldDelim, textInfoToString);
        return kwic;
    }
    
    /**
     * Getter/setter for the delimiter when the input is delimited strings
     * @param value is a string
     */
    this.fieldDelim = function(value) {
        if (!arguments.length) {
            return fldDelim;
        }
        fldDelim = value;
        tmu = textmodel.TextModelUtilities(fldNames, fldDelim, textInfoToString);
        return kwic;
    }
    
    /**
     * Getter/setter for the toString function of TextInfor when the input is delimited strings
     * @param value is a string
     */
    this.textInfoToString = function(value) {
        if (!arguments.length) {
            return textInfoToString;
        }
        textInfoToString = value;
        tmu = textmodel.TextModelUtilities(fldNames, fldDelim, textInfoToString);
        return kwic;
    }
    
    /**
     * Getter/setter for the number of results to show per page
     * @param value is an integer
     * <p>
     * Note: this does <strong>not</strong> redraw the KWIC. Use {@link kwicIS.KWIC#show} for that.
     */
    this.numToShow = function(value) {
        if (!arguments.length) {
            return numToShow;
        }
        numToShow = value;
        calcLastPage();
        return kwic;
    }
    
    /**
     * Getter/setter for the number of the current page of results
     * @param value is an integer
     * <p>
     * Note: this does <strong>not</strong> redraw the KWIC. Use {@link kwicIS.KWIC#show} for that.
     */
    this.whichPage = function(value) {
        if (!arguments.length) {
            return whichPage;
        }
        whichPage = value;
        return kwic;
    }
    
    /**
     * Getter/setter for the length of the left-hand context
     * @param value is an integer
     * <p>
     * Note: this does <strong>not</strong> redraw the KWIC. Use {@link kwicIS.KWIC#show} for that.
     */
    this.leftContextLen = function(value) {
        if (!arguments.length) {
            return leftContextLen;
        }
        leftContextLen = value;
        setupHeaderData();
        return kwic;
    }
    
    /**
     * Getter/setter for the length of the right-hand context
     * @param value is an integer
     * <p>
     * Note: this does <strong>not</strong> redraw the KWIC. Use {@link kwicIS.KWIC#show} for that.
     */
    this.rtContextLen = function(value) {
        if (!arguments.length) {
            return rtContextLen;
        }
        rtContextLen = value;
        setupHeaderData();
        return kwic;
    }
    
    /**
     * Convenience Getter/setter when both contexts are the same length
     * @param value is an integer
     * <p>
     * Note: this does <strong>not</strong> redraw the KWIC. Use {@link kwicIS.KWIC#show} for that.
     */
    this.contextLen = function(value) {
        if (!arguments.length) {
            return rtContextLen;
        }
        leftContextLen = value;
        rtContextLen = value;
        setupHeaderData();
        return kwic;
    }
    
    /**
     * Getter/setter for the script is right to left
     * @param value is a boolean: true means the script is written right to left
     * <p>
     * Note: this does <strong>not</strong> redraw the KWIC. Use {@link kwicIS.KWIC#show} for that.
     */
    this.rtToLeft = function(value) {
        if (!arguments.length) {
            return prefixesOnRight;
        }
        prefixesOnRight = value;
        return kwic;
    }
    
    
    /**
     * @returns the number of hits for these (possibly filtered) results
     */
    this.hitsCount = function() {
        return numHits;
    }
    
     /**
     * Returns the number of hits for these (possibly filtered) results, aggregated by the given field
     * @param field is the field to use for the aggregation
     * @returns a hash of the results
     */
    this.hitsByField = function(field) {
        //return hash of field's values and their counts
        
        var what = {};
        
        data.filter(function(d) {
            return ! excludedIDs.has(d.id);
        }).forEach(function(d) {
            var which = d.hit[field];
            if (typeof(what[which]) === 'undefined') {
                what[which] = 0;
            }
           what[ which ]++;
        });
        return what;
        
    }
    
    /**
     * Returns the actual fields used in the item objects
     * @returns a d3.set
     */
    this.actualFields = function() {
        if (typeof(data) != 'undefined' && data.length > 0) {
            return d3.set(Object.keys(data[0].hit));
        }
        return d3.set();
    }
    
    /**
     * Sets the function to be used for the mouse hover title
     * @param theFun is a function(itemObject, i) and returns a string. <em>i</em> is the number <em>in the current page</em>
     */
    this.cellTitle = function(theFun) {
        cellTitle = theFun;
        return kwic;
    }
    
    /**
     * Sets the function to be called when the column header is clicked
     * @param theFun is a function(headerContent, column)
     */
    this.colHeaderClick = function(theFun) {
        colHeaderClick = theFun;
        return kwic;
    }
    
    /**
     * Sets the function to be called when an item cell is clicked
     * @param theFun is a function(itemObject, i).  <em>i</em> is the number <em>in the current page</em> 
     *
     */
    this.itemCellClick = function(theFun) {
        itemCellClick = theFun;
        return kwic;
    }
    
    //setter only
    //theFun(textInfoArray, hitNumber)
    /**
     * Sets the function to display a string for the row in an extra column, and whether the column is on the right or left end of the row. Default is no extra string.
     * @param theFun is a function(itemObjectArray, hitNumber) and returns a string
     * @param onRight is <em>boolean</em> (not just truthy) for position of rowString &mdash; true is on the right and false is on the left. Default is on the right.
     */
    this.rowString = function(theFun, onRight) {
        rowString = theFun;
        if (typeof(onRight) == "boolean") {
            rowStringOnRight = onRight;
        } else {
            rowStringOnRight = true; //default
        }
        haveRowString = true;
        setupHeaderData();
        return kwic;
    }
    
    
    //other functions
    
    
    /**
     * Sorts the data according to the given column and the relevant settings for field, direction, and order
     * @param column the column to be sorted by
     * @param settings optional object with keys:
     * <ul>
     * <li>field: the field to use as the sort key</li>
     * <li>sortFun: function(a,b) where a and be are the values of <em>field</em> and returns -1,0,1 according to whether a is less than, equal to, or greater than b.
     * Default is a case insensitive lexicographic sort
     * </li>
     * <li>reverse: true if the order of sortFun should be reversed. Default is false.</li>
     * </ul>
     */
    this.sort = function(column, settings) {
        settings = settings || {};
        var field;
        var reverse = false;
        var sortFun;
        
        if (settings.hasOwnProperty('field')) {
            field = settings['field'];
        }
        if (settings.hasOwnProperty('reverse')) {
            reverse = settings['reverse'];
        }
        if (settings.hasOwnProperty('sortFun')) {
            sortFun = settings['sortFun'];
        }
        
        if (typeof(sortFun) === 'undefined') {
            sortFun = function(a,b) {
                var lowerA = a.toLocaleLowerCase();
                var lowerB = b.toLocaleLowerCase();
                //return lowerA < lowerB ? -1 : lowerA > lowerB ? 1 : 0;
                return lowerA.localeCompare(lowerB);
            }
        }
        
        
        var hitCol = leftContextLen +1; //user columns are 1 based
        if (haveRowString && ! rowStringOnRight) {
            column--;
        }
        data.sort(function(obj1, obj2) {
           
           //get the relevant element
            var elt1, elt2;
            if (column == hitCol) {
                elt1 = obj1['hit'];
                elt2 = obj2['hit'];
            } else if (column < hitCol) {
                var which = column-1;
                elt1 = obj1['left'][which];
                elt2 = obj2['left'][which];
            } else  {
                var which = column-(leftContextLen +2); //user columns are 1 based
                elt1 = obj1['right'][which];
                elt2 = obj2['right'][which];
            }
            
            //get the relevant string from the element
            var comp1, comp2;
           
            if (typeof(elt1) === 'undefined') {
                comp1 = "";
            } else {
                if (typeof(field) !== 'undefined') {
                    if (typeof(elt1[field]) === 'undefined') {
                        comp1 = "";
                    } else {
                        comp1 = elt1[field];
                    }
                } else {
                    comp1 = elt1.toString();
                }
            }
            if (typeof(elt2) === 'undefined') {
                comp2 = "";
            } else {
                if (typeof(field) !== 'undefined') {
                    if (typeof(elt2[field]) === 'undefined') {
                        comp2 = "";
                    } else {
                        comp2 = elt2[field];
                    }
                } else {
                    comp2 = elt2.toString();
                }
            }
            
            return sortFun(comp1, comp2);
            
        });
        
        if (reverse) {
            data.reverse();
        }
        
        kwic.showPage(1);
    }
    
    //pageNum is 1-based for users, but internally we use 0-based
    /**
     * Show the given page of results
     * @param pageNum is an integer. Page numbers are 1-based.
     */
    this.showPage = function(pageNum) {
        kwic.whichPage(pageNum -1).show();
    }
    
    /**
     * Show the next page of results
     */
    this.showNextPage = function() {
        if (whichPage == lastPage) {
            return;
        }
        whichPage += 1;
        kwic.show();
    }
    
    /**
     * Show the previous page of results
     */
    this.showPrevPage = function() {
        if (whichPage == 0) {
            return;
        }
        kwic.whichPage( kwic.whichPage()-1 ).show();
    }
    
    /**
     * Show the first page of results
     */
    this.showFirstPage = function() {
        kwic.whichPage(0).show();
    }
    
    /**
     * Show the last page of results
     */
    this.showLastPage = function() {
        kwic.whichPage(lastPage).show();
    }
    
    /**
     * Show the current page of results. This is the function that does the work of showing the KWIC.
     */
    this.show = function() {
        
        //select items for this page, trimming left and rt context, return flattened array
        var startElt = numToShow*whichPage;
        var thisData = data.filter(function(d) {
            return ! excludedIDs.has(d.id);
        })
        .slice(startElt, startElt + numToShow).map(function(d, i){
            d.left = d.left.slice(-leftContextLen);
            //pad if it is too short
            while (d.left.length < leftContextLen) {
                d.left.unshift("");
            }
            d.right = d.right.slice(0, rtContextLen);
            //pad if too short
            while (d.right.length < rtContextLen) {
                d.right.push("");
            }
           
            var what = d.left.concat([d.hit], d.right);
            var rowStr;
            if (! prefixesOnRight) {
                rowStr = [ rowString(what, startElt + i) ];
            } else {
                var tmpWhat = what.slice();
                tmpWhat.reverse();
                rowStr = [ rowString(tmpWhat, startElt + i) ];
            }
            
            
            
            //var rowStr = [ rowString(what, startElt + i) ];
            if (haveRowString) {
                if (rowStringOnRight) {
                    what = what.concat(rowStr);
                } else {
                    what = rowStr.concat(what);
                }
            }
            return {'cells':what, 'index': d.index, 'id':d.id};
        
        });
        
        headers = d3.select(where).select(" table > thead > tr").selectAll("th")
            .data(headerData);
            
        headers.enter()
            .append("th")
            .text(String)
            .classed("kwicTH", true)
            .classed("kwicTHHit", function(d,i){
                return d == "HIT"; //from setupHeaderData    
            })
            .classed("kwicRowSummary", function(d,i) {
                return isSummaryCol(i);
            })
            .on("click", colHeaderClick);
        
        
        /* https://github.com/d3/d3/blob/master/CHANGES.md#selections-d3-selection
        var circle = svg.selectAll("circle").data(data) // UPDATE
			.style("fill", "blue");

		circle.exit().remove(); // EXIT

		circle.enter().append("circle") // ENTER
			.style("fill", "green")
		  .merge(circle) // ENTER + UPDATE
			.style("stroke", "black");
    
	    */
	   
	    
	    //tr
	    //UPDATE
        var tr = d3.select(where).select(" table > tbody").selectAll("tr")
            .data(thisData);
            
        //EXIT
        tr.exit().remove();
        
        var tre = tr.enter().append("tr") //ENTER
         .merge(tr) //ENTER + UPDATE
            .attr("index", function(d) {
            	return d.index;
        });
        

        //td
       	//UPDATE
        var td = tre.selectAll("td")
            .data(function(d) {
                return d.cells;
            });
            
        td.enter().append("td")	//ENTER
            .on("click", itemCellClick)
         .merge(td) //ENTER + UPDATE
            //td.html(function(d) {
            .html(function(d) {
                return d.toString()})
            .attr("class","kwicTD")
            .classed("kwicTDHit", function(d,i) {
                var isHit;
                if (!haveRowString || rowStringOnRight) {
                    isHit = (i == leftContextLen);
                } else if (haveRowString) {
                    isHit = (i == (1+ leftContextLen));
                }
                return isHit;    
            })
            .classed("kwicRowSummary", function(d,i) {
                return isSummaryCol(i);
            })
            .attr("title", cellTitle);


        //pass the hit's id and index from the row to the cell
        d3.select(where).selectAll('.kwicTDHit')
            .each(function(d) {
               var p = d3.select(this.parentNode);
                
                d.id = +p.attr('id');
                d.index = p.attr('index');
                
                return d;
            });
            
        
        //EXIT
        //td.exit().remove(); //we don't need this since tr.exit() removes the whole row
        
        return;
    }
    
    //convenience filter function factories 

    /**
     * Convenience filter function factory for including items whose field is val
     * @param field is the field to use
     * @param val is the value to test
     * @returns a function(itemObject) which returns true if itemObject[field] === val and false otherwise
     */
    this.include_filter_function = function(field, val) {
        return function(d) {
            return d[field] === val;
        }
    }
    
    //return a function that includes an item where the field does NOT have the value
     /**
     * Convenience filter function factory for excluding items whose field is val
     * @param field is the field to use
     * @param val is the value to test
     * @returns a function(itemObject) which returns true if itemObject[field] !== val and false otherwise
     */
    this.exclude_filter_function = function(field, val) {
        return function(d) {
            return d[field] !== val;
        }
    }
    
    ///////////     private functions
    
    function setupHeaderData() {
        //set up the headers
        headerData = [];
        for(var i=0;i<leftContextLen;i++) {
            if (! prefixesOnRight) {
                headerData.push("-" + (leftContextLen -i))
            } else {
                headerData.push("+" + (leftContextLen -i))
            }
            
        }
        headerData.push("HIT");
        for(var i=0;i<rtContextLen;i++) {
            if (!prefixesOnRight) {
                headerData.push("+" + (1+i))
            } else {
                headerData.push("-" + (1+i))
            }
            
        }
        if (haveRowString) {
            if (rowStringOnRight) {
                headerData.push("Summary");
            } else {
                headerData.unshift("Summary");
            }    
        }
        
    }
    
    function isSummaryCol(i) {
        if (haveRowString) {
            if (rowStringOnRight) {
                return i == headers.size()-1;
            } else if (i == 0) {
                return true;
            }
        }
        return false;
    }
    
    function calcLastPage() {
        numHits = data.length - excludedIDs.size();
        if (!data || numHits === 0) {
            lastPage = 0;
        } else {
            lastPage = Math.floor((numHits-1) / numToShow) ;
        }
    }
    
    function filter() {
        excludedIDs = d3.set();
        
        
        var hitHasFilter = typeof(filters.hit) === 'function';
        var haveFilter = hitHasFilter;
        
        var leftHasFilter = [];
        for (var i=0, n=data[0].left.length; i<n; i++) {
            leftHasFilter[i] = typeof(filters.left) !== 'undefined' && typeof(filters.left[i]) === 'function';
            haveFilter = haveFilter || leftHasFilter[i];
        }
        
        var rightHasFilter = [];
        for (var i=0, n=data[0].left.length; i<n; i++) {
            rightHasFilter[i] = typeof(filters.right) !== 'undefined' && typeof(filters.right[i]) === 'function';
            haveFilter = haveFilter || rightHasFilter[i];
        }
        
        //shortcut if we have no filters
        if (! haveFilter) {
            calcLastPage();
            return;
        }
        
        for(var i=0, n=data.length;i<n;i++) {
            var d = data[i];
            
            if (hitHasFilter && ! filters.hit(d.hit)) {
                excludedIDs.add(d.id);
                continue;
            }
            
            var foundOne = false;
            for(var j=0, m=d.left.length; j<m; j++) {
                if (leftHasFilter[j] && ! filters.left[j](d.left[m-j-1])) {
                    excludedIDs.add(d.id);
                    foundOne = true;
                    break;
                }
            }
            if (foundOne) {
                continue;
            }
            
            for(var j=0, m=d.right.length; j<m; j++) {
                if (rightHasFilter[j] && ! filters.right[j](d.right[j])) {
                    excludedIDs.add(d.id);
                    break;
                }
            }
            
        }
        calcLastPage();
    }
    
}

})()