/* (This is the new BSD license.)
* Copyright (c) 2014, Chris Culy
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the Chris Culy nor the 
*		names of its contributors may be used to endorse or promote 
*		products from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY Chris Culy
* ``AS IS'' AND ANY OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
* THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL Chris Culy
* BE LIABLE FOR ANY, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
* GOODS OR SERVICES; OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
* CAUSED AND ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR 
* TORT INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
* THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

"use strict";
/**
 @namespace textmodel
 All of the functionality is in the textmodel namespace
*/
var textmodel = textmodel || {};

(function(){
//split string on spaces (assume already tokenized/analyzed)
 /**
  * @class textmodel.RegExTextModel
  *  This is the class for the RegExTextModel document model.
  * <p>
  *  
  */
textmodel.RegExTextModel = function(string) {
    var that = this;
    var s = string;
    
    var aveItemLen = 20; //will use to look for context
    
    
    var initialWhitespace = /^\s/;
    var finalWhitespace = /\s$/;
    
    /**
     * @param re is the search expression. Only the <em>i</em> and <em>m</em> flags may be used. g may not.
     * @param contextLen the length of the preceding and following context to be returned
     * @param maxRandomHits how many random hits to return. -1 or null to return all
     * @returns array of [array of prefixes, array of item, array of suffixes, array of ids],
     * where prefixes and suffixes come from splitting on whitespace, and the item is extended in both directions to the nearest whitespace
     */
    this.getItem = function(re, contextLen, maxRandomHits) {
        var prefixArray = [], itemArray = [], suffixArray = [], idArray = [];
       
        if (typeof(re.source) === 'undefined' || re.source === "") {
            return [prefixArray, itemArray, suffixArray, idArray];
        }
       
        var flags = "g";
        if (re.ignoreCase) {
            flags += "i";
        }
        if (re.multiline) {
            flags += "m";
        }
        var regex = new RegExp(re.source, flags);
        
        
        //based on https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/RegExp/exec
        var results = [];
        var oneResult = [];
        while ((oneResult = regex.exec(s)) !== null) {
            var result = {'found' : oneResult[0]};
            result.start = oneResult.index;
            result.end = regex.lastIndex; //this is index just past the end of the match
            results.push(result);
        }
        
        //select the number of random hits
        if (maxRandomHits !== null && maxRandomHits > 0) {
            //create an array 0..n, shuffle it, then take the first maxRandomHits
            var indices = d3.shuffle(d3.range(0,results.length)).slice(0,maxRandomHits);
            
            //now permute our original
            results = d3.permute(results, indices);
        }
        
        //now get the contexts and fill our arrays
        
        
        for (var i=0, n=results.length;i<n;i++) {
            var r = results[i];
            
            var hit = r.found;
            
            //if the hit is internal to an item, then we need to expand it to the nearest whitespace, in either direction
            
            var expandLeft = (! initialWhitespace.test(hit));
            if (expandLeft && initialWhitespace.test(s.substring(r.start-1,r.start))) {
                expandLeft = false;
            }
            var expandRt = (! finalWhitespace.test(hit))
            if (expandRt && initialWhitespace.test(s.substring(r.end,r.end+1))) {
                expandRt = false;
            }
            
            
            var thisContextLen = contextLen;
            if (expandLeft || expandRt) {
                thisContextLen += 1;
            }
            var padding = (thisContextLen+2)*aveItemLen;
            
            var pieces = s.substring( r.start-padding-1, r.start ).split(/\s+/);
            if (pieces[pieces.length-1] === '') {
                pieces.pop();
            }
            if (expandLeft && pieces.length > 0) {
              hit = pieces.pop() + hit;
            }
            prefixArray[i] = pieces.slice(-contextLen);
            
            pieces = s.substring( r.end, r.end+padding ).split(/\s+/);
            if (pieces[0] === '') {
                pieces.shift();
            }
            if (expandRt && pieces.length > 0) {
              hit += pieces.shift();
            }
            suffixArray[i] = pieces.slice(0,contextLen);
            
            
            itemArray[i] = hit;
            idArray[i] = r.start;
            
        }
        
        return [prefixArray, itemArray, suffixArray, idArray];
    }
    
    /**
     * Get a previously calculated item at the index along with some context on either side
     * @param re is a the regular expression used previsousy by {@link textmodel.RegExTextModel.KWIC#getItem} to calculate the index
     * @param index is an index as calculated by {@link textmodel.RegExTextModel.KWIC#getItem}
     * @param contextLen is the length (in characters) of the context
     * @returns an object  {item:String, left:String, right:String}
     */
    this.getItemAndContext = function(re, index, contextLen) {
        index = 1*index;
        var what = {};
        var flags = "";
        if (re.ignoreCase) {
            flags += "i";
        }
        if (re.multiline) {
            flags += "m";
        }
        var regex = new RegExp(re.source, flags);
        
        var r = regex.exec(s.substr(index));
        
        
        
        if (! r[0]) {
            return null; //error
        }
        
        what.item = r[0];
        var beginL = index-contextLen;
        var realLen = contextLen;
        if (index < contextLen) {
            beginL = 0;
            realLen = index;
        }
        what.left = s.substr(beginL, realLen);
        what.right = s.substr(index + what.item.length, contextLen)
        
        return what;
    }
    
}
})();
